@@ -28,7 +28,7 @@ import time
 from urllib import parse
 import uuid
 
-from amqp import exceptions as amqp_exec
+from amqp import exceptions as amqp_ex
 import kombu
 import kombu.connection
 import kombu.entity
@@ -165,9 +165,16 @@ rabbit_opts = [
                 'Raft consensus algorithm. It is available as of '
                 'RabbitMQ 3.8.0. If set this option will conflict with '
                 'the HA queues (``rabbit_ha_queues``) aka mirrored queues, '
-                'in other words the HA queues should be disabled, quorum '
-                'queues durable by default so the amqp_durable_queues '
-                'opion is ignored when this option enabled.'),
+                'in other words the HA queues should be disabled. '
+                'Quorum queues are also durable by default so the '
+                'amqp_durable_queues option is ignored when this option is '
+                'enabled.'),
+    cfg.BoolOpt('rabbit_transient_quorum_queue',
+                default=False,
+                help='Use quorum queues for transients queues in RabbitMQ. '
+                'Enabling this option will then make sure those queues are '
+                'also using quorum kind of rabbit queues, which are HA by '
+                'default.'),
     cfg.IntOpt('rabbit_quorum_delivery_limit',
                default=0,
                help='Each time a message is redelivered to a consumer, '
@@ -195,12 +202,16 @@ rabbit_opts = [
                'Used only when rabbit_quorum_queue is enabled, '
                'Default 0 which means dont set a limit.'),
     cfg.IntOpt('rabbit_transient_queues_ttl',
-               min=1,
+               min=0,
                default=1800,
                help='Positive integer representing duration in seconds for '
                     'queue TTL (x-expires). Queues which are unused for the '
                     'duration of the TTL are automatically deleted. The '
-                    'parameter affects only reply and fanout queues.'),
+                    'parameter affects only reply and fanout queues. Setting '
+                    '0 as value will disable the x-expires. If doing so, '
+                    'make sure you have a rabbitmq policy to delete the '
+                    'queues or you deployment will create an infinite number '
+                    'of queue over time.'),
     cfg.IntOpt('rabbit_qos_prefetch_count',
                default=0,
                help='Specifies the number of messages to prefetch. Setting to '
@@ -211,7 +222,7 @@ rabbit_opts = [
                "considered down if heartbeat's keep-alive fails "
                "(0 disables heartbeat)."),
     cfg.IntOpt('heartbeat_rate',
-               default=2,
+               default=3,
                help='How often times during the heartbeat_timeout_threshold '
                'we check the heartbeat.'),
     cfg.BoolOpt('direct_mandatory_flag',
@@ -230,7 +241,24 @@ rabbit_opts = [
                 default=False,
                 help="Enable x-cancel-on-ha-failover flag so that "
                      "rabbitmq server will cancel and notify consumers"
-                     "when queue is down")
+                     "when queue is down"),
+    cfg.BoolOpt('use_queue_manager',
+                default=False,
+                help='Should we use consistant queue names or random ones'),
+    cfg.StrOpt('hostname',
+               default=socket.gethostname(),
+               help='Hostname used by queue manager'),
+    cfg.StrOpt('processname',
+               default=os.path.basename(sys.argv[0]),
+               help='Process name used by queue manager'),
+    cfg.BoolOpt('rabbit_stream_fanout',
+                default=False,
+                help='Use stream queues in RabbitMQ (x-queue-type: stream). '
+                'Streams are a new persistent and replicated data structure '
+                '("queue type") in RabbitMQ which models an append-only log '
+                'with non-destructive consumer semantics. It is available '
+                'as of RabbitMQ 3.9.0. If set this option will replace all '
+                'fanout queues with only one stream queue.'),
 ]
 
 LOG = logging.getLogger(__name__)
@@ -238,7 +266,8 @@ LOG = logging.getLogger(__name__)
 
 def _get_queue_arguments(rabbit_ha_queues, rabbit_queue_ttl,
                          rabbit_quorum_queue,
-                         rabbit_quorum_queue_config):
+                         rabbit_quorum_queue_config,
+                         rabbit_stream_fanout):
     """Construct the arguments for declaring a queue.
 
     If the rabbit_ha_queues option is set, we try to declare a mirrored queue
@@ -275,11 +304,15 @@ def _get_queue_arguments(rabbit_ha_queues, rabbit_queue_ttl,
 
     rabbit_quorum_queue_config:
     Quorum queues provides three options to handle message poisoning
-    and limit the resources the qourum queue can use
+    and limit the resources the quorum queue can use
     x-delivery-limit number of times the queue will try to deliver
     a message before it decide to discard it
     x-max-in-memory-length, x-max-in-memory-bytes control the size
     of memory used by quorum queue
+
+    If the rabbit_stream_fanout option is set, fanout queues are going to use
+    stream instead of quorum queues. See here:
+      https://www.rabbitmq.com/streams.html
     """
     args = {}
 
@@ -306,6 +339,12 @@ def _get_queue_arguments(rabbit_ha_queues, rabbit_queue_ttl,
     if rabbit_queue_ttl > 0:
         args['x-expires'] = rabbit_queue_ttl * 1000
 
+    if rabbit_stream_fanout:
+        args = {'x-queue-type': 'stream'}
+        if rabbit_queue_ttl > 0:
+            # max-age is a string
+            args['x-max-age'] = f"{rabbit_queue_ttl}s"
+
     return args
 
 
@@ -332,7 +371,8 @@ class Consumer(object):
                  exchange_auto_delete, queue_auto_delete, callback,
                  nowait=False, rabbit_ha_queues=None, rabbit_queue_ttl=0,
                  enable_cancel_on_failover=False, rabbit_quorum_queue=False,
-                 rabbit_quorum_queue_config=QuorumMemConfig(0, 0, 0)):
+                 rabbit_quorum_queue_config=QuorumMemConfig(0, 0, 0),
+                 rabbit_stream_fanout=False):
         """Init the Consumer class with the exchange_name, routing_key,
         type, durable auto_delete
         """
@@ -348,7 +388,7 @@ class Consumer(object):
         rabbit_quorum_queue_config = rabbit_quorum_queue_config or {}
         self.queue_arguments = _get_queue_arguments(
             rabbit_ha_queues, rabbit_queue_ttl, rabbit_quorum_queue,
-            rabbit_quorum_queue_config)
+            rabbit_quorum_queue_config, rabbit_stream_fanout)
         self.queue = None
         self._declared_on = None
         self.exchange = kombu.entity.Exchange(
@@ -415,7 +455,7 @@ class Consumer(object):
                       conn.connection_id, self.queue_name)
             try:
                 self.queue.declare()
-            except amqp_exec.PreconditionFailed as err:
+            except amqp_ex.PreconditionFailed as err:
                 # NOTE(hberaud): This kind of exception may be triggered
                 # when a control exchange is shared between services and
                 # when services try to create it with configs that differ
@@ -424,6 +464,29 @@ class Consumer(object):
                 # from the one used first.
                 LOG.warning(err)
                 self._declare_fallback(err, conn, consumer_arguments)
+            except amqp_ex.NotFound as ex:
+                # NOTE(viktor.krivak): This exception is raised when
+                # non-durable and non-ha queue is hosted on node that
+                # is currently unresponsive or down. Thus, it is
+                # not possible to redeclare the queue since its status
+                # is unknown, and it could disappear in
+                # a few moments. However, the queue could be explicitly
+                # deleted and the redeclaration will then succeed.
+                # Queue will be then scheduled on any of the
+                # running/responsive nodes.
+                # This fixes bug:
+                # https://bugs.launchpad.net/oslo.messaging/+bug/2068630
+
+                LOG.warning("Queue %s is stuck on unresponsive node. "
+                            "Trying to delete the queue and redeclare it "
+                            "again, Error info: %s", self.queue_name, ex)
+                try:
+                    self.queue.delete()
+                except Exception as in_ex:
+                    LOG.warning("During cleanup of stuck queue deletion "
+                                "another exception occurred: %s. Ignoring...",
+                                in_ex)
+                self.queue.declare()
 
         except conn.connection.channel_errors as exc:
             # NOTE(jrosenboom): This exception may be triggered by a race
@@ -453,6 +516,14 @@ class Consumer(object):
                           'Queue: [%(queue)s], '
                           'error message: [%(err_str)s]', info)
                 time.sleep(interval)
+                if self.queue_arguments.get('x-queue-type') == 'quorum':
+                    # Before re-declare queue, try to delete it
+                    # This is helping with issue #2028384
+                    # NOTE(amorin) we need to make sure the connection is
+                    # established again, because when an error occur, the
+                    # connection is closed.
+                    conn.ensure_connection()
+                    self.queue.delete()
                 self.queue.declare()
             else:
                 raise
@@ -495,6 +566,24 @@ class Consumer(object):
                                    nowait=self.nowait)
             else:
                 raise
+        except amqp_ex.InternalError as exc:
+            if self.queue_arguments.get('x-queue-type') == 'quorum':
+                # Before re-consume queue, try to delete it
+                # This is helping with issue #2028384
+                if exc.code == 541:
+                    LOG.warning('Queue %s seems broken, will try delete it '
+                                'before starting over.', self.queue.name)
+                    # NOTE(amorin) we need to make sure the connection is
+                    # established again, because when an error occur, the
+                    # connection is closed.
+                    conn.ensure_connection()
+                    self.queue.delete()
+                    self.declare(conn)
+                    self.queue.consume(callback=self._callback,
+                                       consumer_tag=str(tag),
+                                       nowait=self.nowait)
+            else:
+                raise
 
     def cancel(self, tag):
         LOG.trace('ConsumerBase.cancel: canceling %s', tag)
@@ -609,6 +698,9 @@ class Connection(object):
         self.rabbit_quorum_queue = driver_conf.rabbit_quorum_queue
         self.rabbit_quorum_queue_config = self._get_quorum_configurations(
             driver_conf)
+        self.rabbit_transient_quorum_queue = \
+            driver_conf.rabbit_transient_quorum_queue
+        self.rabbit_stream_fanout = driver_conf.rabbit_stream_fanout
         self.rabbit_transient_queues_ttl = \
             driver_conf.rabbit_transient_queues_ttl
         self.rabbit_qos_prefetch_count = driver_conf.rabbit_qos_prefetch_count
@@ -626,6 +718,18 @@ class Connection(object):
         self.heartbeat_in_pthread = driver_conf.heartbeat_in_pthread
         self.ssl_enforce_fips_mode = driver_conf.ssl_enforce_fips_mode
         self.enable_cancel_on_failover = driver_conf.enable_cancel_on_failover
+        self.use_queue_manager = driver_conf.use_queue_manager
+
+        if self.rabbit_stream_fanout and self.rabbit_qos_prefetch_count <= 0:
+            raise RuntimeError('Configuration Error: rabbit_stream_fanout '
+                               'need rabbit_qos_prefetch_count to be set to '
+                               'a value greater than 0.')
+
+        if (self.rabbit_stream_fanout and not
+                self.rabbit_transient_quorum_queue):
+            raise RuntimeError('Configuration Error: rabbit_stream_fanout '
+                               'need rabbit_transient_quorum_queue to be set '
+                               'to true.')
 
         if self.heartbeat_in_pthread:
             # NOTE(hberaud): Experimental: threading module is in use to run
@@ -805,6 +909,13 @@ class Connection(object):
             self.connection.port = 1234
             self._poll_timeout = 0.05
 
+        if self.use_queue_manager:
+            self._q_manager = amqpdriver.QManager(
+                hostname=driver_conf.hostname,
+                processname=driver_conf.processname)
+        else:
+            self._q_manager = None
+
     # FIXME(markmc): use oslo sslutils when it is available as a library
     _SSL_PROTOCOLS = {
         "tlsv1": ssl.PROTOCOL_TLSv1,
@@ -1065,11 +1176,20 @@ class Connection(object):
         """Close/release this connection."""
         self._heartbeat_stop()
         if self.connection:
-            for consumer in filter(lambda c: c.type == 'fanout',
-                                   self._consumers):
-                LOG.debug('[connection close] Deleting fanout '
-                          'queue: %s ' % consumer.queue.name)
-                consumer.queue.delete()
+            # NOTE(jcosmao) Delete queue should be called only when queue name
+            # is randomized. When using streams, queue is shared between
+            # all consumers, thus deleting fanout queue will force all other
+            # consumers to disconnect/reconnect by throwing
+            # amqp.exceptions.ConsumerCancelled.
+            # When using QManager, queue name is consistent accross agent
+            # restart, so we don't need to delete it either. Deletion must be
+            # handled by expiration policy.
+            if not self.rabbit_stream_fanout and not self.use_queue_manager:
+                for consumer in filter(lambda c: c.type == 'fanout',
+                                       self._consumers):
+                    LOG.debug('[connection close] Deleting fanout '
+                              'queue: %s ' % consumer.queue.name)
+                    consumer.queue.delete()
             self._set_current_channel(None)
             self.connection.release()
             self.connection = None
@@ -1204,7 +1324,7 @@ class Connection(object):
                             ConnectionRefusedError,
                             OSError,
                             kombu.exceptions.OperationalError,
-                            amqp_exec.ConnectionForced) as exc:
+                            amqp_ex.ConnectionForced) as exc:
                         LOG.info("A recoverable connection/channel error "
                                  "occurred, trying to reconnect: %s", exc)
                         self.ensure_connection()
@@ -1315,13 +1435,15 @@ class Connection(object):
             queue_name=topic,
             routing_key='',
             type='direct',
-            durable=False,
+            durable=self.rabbit_transient_quorum_queue,
             exchange_auto_delete=False,
             queue_auto_delete=False,
             callback=callback,
             rabbit_ha_queues=self.rabbit_ha_queues,
             rabbit_queue_ttl=self.rabbit_transient_queues_ttl,
-            enable_cancel_on_failover=self.enable_cancel_on_failover)
+            enable_cancel_on_failover=self.enable_cancel_on_failover,
+            rabbit_quorum_queue=self.rabbit_transient_quorum_queue,
+            rabbit_quorum_queue_config=self.rabbit_quorum_queue_config)
 
         self.declare_consumer(consumer)
 
@@ -1347,22 +1469,35 @@ class Connection(object):
     def declare_fanout_consumer(self, topic, callback):
         """Create a 'fanout' consumer."""
 
-        unique = uuid.uuid4().hex
         exchange_name = '%s_fanout' % topic
-        queue_name = '%s_fanout_%s' % (topic, unique)
+        if self.rabbit_stream_fanout:
+            queue_name = '%s_fanout' % topic
+        else:
+            if self._q_manager:
+                unique = self._q_manager.get()
+            else:
+                unique = uuid.uuid4().hex
+            queue_name = '%s_fanout_%s' % (topic, unique)
+        LOG.debug('Creating fanout queue: %s', queue_name)
+
+        is_durable = (self.rabbit_transient_quorum_queue or
+                      self.rabbit_stream_fanout)
 
         consumer = Consumer(
             exchange_name=exchange_name,
             queue_name=queue_name,
             routing_key=topic,
             type='fanout',
-            durable=False,
+            durable=is_durable,
             exchange_auto_delete=True,
             queue_auto_delete=False,
             callback=callback,
             rabbit_ha_queues=self.rabbit_ha_queues,
             rabbit_queue_ttl=self.rabbit_transient_queues_ttl,
-            enable_cancel_on_failover=self.enable_cancel_on_failover)
+            enable_cancel_on_failover=self.enable_cancel_on_failover,
+            rabbit_quorum_queue=self.rabbit_transient_quorum_queue,
+            rabbit_quorum_queue_config=self.rabbit_quorum_queue_config,
+            rabbit_stream_fanout=self.rabbit_stream_fanout)
 
         self.declare_consumer(consumer)
 
@@ -1406,7 +1541,7 @@ class Connection(object):
         if not (exchange.passive or exchange.name in self._declared_exchanges):
             try:
                 exchange(self.channel).declare()
-            except amqp_exec.PreconditionFailed as err:
+            except amqp_ex.PreconditionFailed as err:
                 # NOTE(hberaud): This kind of exception may be triggered
                 # when a control exchange is shared between services and
                 # when services try to create it with configs that differ
@@ -1469,7 +1604,8 @@ class Connection(object):
                     self.rabbit_ha_queues,
                     0,
                     self.rabbit_quorum_queue,
-                    self.rabbit_quorum_queue_config))
+                    self.rabbit_quorum_queue_config,
+                    False))
             log_info = {'key': routing_key, 'exchange': exchange}
             LOG.trace(
                 'Connection._publish_and_creates_default_queue: '
@@ -1508,13 +1644,16 @@ class Connection(object):
 
     def direct_send(self, msg_id, msg):
         """Send a 'direct' message."""
-        exchange = kombu.entity.Exchange(name='',  # using default exchange
-                                         type='direct',
-                                         durable=False,
-                                         auto_delete=True,
-                                         passive=True)
+        exchange = kombu.entity.Exchange(
+            name='',  # using default exchange
+            type='direct',
+            durable=self.rabbit_transient_quorum_queue,
+            auto_delete=True,
+            passive=True)
         options = oslo_messaging.TransportOptions(
             at_least_once=self.direct_mandatory_flag)
+
+        LOG.debug('Sending direct to %s', msg_id)
         self._ensure_publishing(self._publish_and_raises_on_missing_exchange,
                                 exchange, msg, routing_key=msg_id,
                                 transport_options=options)
@@ -1528,6 +1667,8 @@ class Connection(object):
             durable=self.durable,
             auto_delete=self.amqp_auto_delete)
 
+        LOG.debug('Sending topic to %s with routing_key %s', exchange_name,
+                  topic)
         self._ensure_publishing(self._publish, exchange, msg,
                                 routing_key=topic, timeout=timeout,
                                 retry=retry,
@@ -1535,11 +1676,13 @@ class Connection(object):
 
     def fanout_send(self, topic, msg, retry=None):
         """Send a 'fanout' message."""
-        exchange = kombu.entity.Exchange(name='%s_fanout' % topic,
-                                         type='fanout',
-                                         durable=False,
-                                         auto_delete=True)
+        exchange = kombu.entity.Exchange(
+            name='%s_fanout' % topic,
+            type='fanout',
+            durable=self.rabbit_transient_quorum_queue,
+            auto_delete=True)
 
+        LOG.debug('Sending fanout to %s_fanout', topic)
         self._ensure_publishing(self._publish, exchange, msg, retry=retry)
 
     def notify_send(self, exchange_name, topic, msg, retry=None, **kwargs):
